/* This BSP program solves the 0-1 Knapsack problem using Branch
 * and Bound 
 * See "Knapsack Problems. Algorithms and Computer Implementations." 
 * by Silvano Martello and Paolo Toth. John Wiley and Sons.
 * This is approach focuses in the aspects of general
 * parallel branch and bound and not in the concrete 
 * features of knapsack problems.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
*/

#include "bsp.h"
#include <stdio.h>
#include <stdlib.h>

#define MAX 4096
#define MAXHEAPSIZE 16
#define STRIDE 1
#define  INFINITY 8192

typedef int number;
typedef int boolean;

typedef struct {
	int k, C, P;
} problemt;

#define ns "%d\n"

number N, M;
number w[MAX], p[MAX];
number bestSol = -INFINITY;

problemt problems[MAXHEAPSIZE];
number head = 0, tail = 0;

FILE *in;

void readKnap(void) {
  number i,j;

	bsp_push_reg(&N,sizeof(int));
	bsp_push_reg(&M,sizeof(int));
	bsp_push_reg(&w,MAX*sizeof(int));
	bsp_push_reg(&p,MAX*sizeof(int));
	bsp_push_reg(&bestSol,sizeof(int));
	bsp_sync();

	if (bsp_pid() == 0) {
		if ((in = fopen("knap.dat","r+")) == NULL) {
		 bsp_abort("Cannot open input file.\n");
		}
		fscanf(in,ns,&N);
		fscanf(in,ns,&M);
	}
	bsp_bcast(0,&N,&N,sizeof(int));
	bsp_bcast(0,&M,&M,sizeof(int));
	for(j=0; j < bsp_nprocs();j++) {
		if (bsp_pid() == j) 
			printf("%d: N = %d M = %d\n",bsp_pid(),N,M);
		bsp_sync();
	}
			
	bsp_sync();
	if (bsp_pid() == 0) {
		for(i=0;i<N;i++) {
			fscanf(in,ns,&w[i]);
			fscanf(in,ns,&p[i]);
		}
		w[N] = INFINITY; p[N] = 0; /* sentinel object */
		fclose(in);
	}
	bsp_sync();
	bsp_bcast(0,w,w,(N+1)*sizeof(int));
	bsp_bcast(0,p,p,(N+1)*sizeof(int));
	for(j=0; j < bsp_nprocs();j++) {
		if (bsp_pid() == j) {
			for(i=0;i<N;i++) {
				if (i%STRIDE == 0) {
					printf("%d: w = %d p = %d\n",bsp_pid(),w[i],p[i]);
				}  
			}
		}
		bsp_sync();
	}
}

void lowerUpper(problemt Prob, number *L, number *U) {
  /* naive approach */
  number i, weig = 0, prof = Prob.P;

  if (Prob.C < 0) {
   *L = -INFINITY; *U = -INFINITY;
  }
  else {
   for(i=Prob.k; weig <= Prob.C; i++) {
    weig += w[i]; prof += p[i];
   }
   /* invariant:  weig > prob.C */
   i--;
   weig -= w[i];
   prof -= p[i];
   *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
   *U = prof+(p[i]*(Prob.C-weig))/w[i];
  }
}


void setProblem(problemt *Prob, number k, number C, number P) {

  Prob->k = k;
  Prob->C = C;
  Prob->P = P;
}

int getProblem(problemt *Prob) {

  if (head != tail)  {
		setProblem(Prob, problems[head].k, problems[head].C, problems[head].P);
		head = (head+1)% MAXHEAPSIZE;
		return 1;
	}
	else return 0;
}


void insertProblem(problemt Prob) {
  number next;

  next = (tail+1)%MAXHEAPSIZE;

  if (next == head) {
   printf("\nError. The heap of problems is full.");
   bsp_abort("\nIncrease the value of MAXHEAPSIZE. Current value is (%d)\n",MAXHEAPSIZE);
  }
  setProblem(&problems[tail],Prob.k, Prob.C, Prob.P);
  tail = next;
}

number numProblems(void) { /* number of problems in the queue */
  return (bsp_sumI((tail-head)+MAXHEAPSIZE)%MAXHEAPSIZE);
}

number numChildren(problemt Prob) {
  return ((Prob.k < N) ? 2:0);
}


void child(problemt father,number order, problemt *child) {

  setProblem(child,
         father.k+1,
         father.C-(1-order)*w[father.k],
         father.P+(1-order)*p[father.k]
        );
}

void branchAndBound(void) {
  problemt Prob, Probk;
  number nC, k, L, U;

  while (numProblems() > 0) {
   if (getProblem(&Prob)) {
		 nC = numChildren(Prob);
		 if (nC != 0) {
			lowerUpper(Prob, &L, &U);
			if (bestSol < L) bestSol = L; /* L <= bestSol <= U */
			if (bestSol < U) {
				for(k=0; k<nC; k++) {
				child(Prob,k,&Probk);
				insertProblem(Probk);
				}
			}
		 } /* nC != 0 */
	 }
	} /* end while */
}

int main(void) {
  int nprocs;
  problemt Prob;

	bsp_begin(bsp_nprocs());
  readKnap();
	if (bsp_pid() == 0) {
		setProblem(&Prob, 0, M, 0);
		insertProblem(Prob);
	}
  branchAndBound();
  printf("\n%d: sol = %d\n", bsp_pid(), bestSol); 
	bsp_sync();
	bsp_end();
}
